/*
     This file is part of GNUnet.
     (C) 2001, 2002 Christian Grothoff (and other contributing authors)

     GNUnet is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     GNUnet is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with GNUnet; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/

/** 
 * @file applications/afs/encoding/requestmanager.c
 * @author Christian Grothoff
 *
 * The RequestManager keeps track and re-issues requests for
 * Content-Nodes.
 **/ 

#include "config.h"

#include <errno.h>
#include <math.h>
#include <string.h>
#include <unistd.h>

#include "util/checksum.h"
#include "util/logging.h"
#include "util/tcp_return.h"
#include "util/xmalloc.h"
#include "util/returnvalues.h"
#include "applications/afs/util/policy.h"
#include "applications/afs/encoding/requestmanager.h"
#include "applications/afs/protocol/messages.h"
#include "util/port.h"

static unsigned int connectedPeers = 0;

static void queryPeerCount(GNUNET_TCP_SOCKET * sock) {
  CS_HEADER q;

  q.size = htons(sizeof(CS_HEADER));
  q.tcpType = htons(CS_PROTO_CLIENT_COUNT);
  if (OK == writeToSocket(sock, &q)) {
    if (SYSERR == readTCPResult(sock, &connectedPeers))
      LOG(LOG_WARNING,
	  "WARNING: gnunetd failed to respond to query for number of connected peers\n");
  } else
    LOG(LOG_WARNING,
	"WARNING: could not send message to gnuentd via TCP\n");
  if (connectedPeers < 10)
    connectedPeers = 10;
}

/**
 * how often do we try a query  (in a row, it will be
 * re-scheduled later!) 
 **/
#define MAX_TRIES 50

/**
 * For debugging, print a RequestEntry to LOG.
 **/
static void printRequestEntry(char * desc,
			      int i,
			      RequestEntry * re) {
  HexName hex;

  IFLOG(LOG_DEBUG,
	hash2hex(&re->message->queries[0], 
		 &hex));
  LOG(LOG_DEBUG,
      "DEBUG: %4i: %s is %s for node %d (%d tries)\n",
      i, 
      desc, 
      &hex, 
      re->receiverNode,
      re->tries);
}

/**
 * Print the contents of the request manager. For debugging.
 **/
void printRequestManager(RequestManager * this) {
  int i;
  MUTEX_LOCK(&this->lock);
  LOG(LOG_DEBUG,
      "DEBUG: RM TTL %u duplicates %d\n",
      this->initialTTL, this->duplicationEstimate);
  for (i=0;i<this->requestedListIndex;i++) 
    printRequestEntry("PENDING", i, this->requestedList[i]);
  for (i=0;i<this->requestListIndex;i++)
    printRequestEntry("WAITING", i, this->requestList[i]);    
  
  MUTEX_UNLOCK(&this->lock);
}

/**
 * Send the request from the requestedList[requestedIndex] out onto
 * the network. If the request has been pending for too long, put it
 * back into the requestList and move another one in its place
 * instead.
 *
 * @param this the RequestManager
 * @param requestedIndex the index of the Request to issue
 **/
static void issueRequest(RequestManager * this,
			 int requestedIndex) {
  RequestEntry * entry;
  cron_t now;
  int ttl;
  int i;
  HexName hex;
  CS_HEADER * msg;
  GNUNET_TCP_SOCKET * sock;

  cronTime(&now);
  entry = this->requestedList[requestedIndex];
  if ( ((entry->tries % MAX_TRIES) == (MAX_TRIES-1)) && 
       (this->requestListIndex > 0) ) { /* swap against an entry
					   from the request list */
    i = randomi(this->requestListIndex);
    this->requestedList[requestedIndex] 
      = this->requestList[i];
    this->requestList[i] = entry;
    entry = this->requestedList[requestedIndex];
  }
  if (entry->lasttime == 0) {
    ttl = this->initialTTL;
  } else {
    ttl = ntohl(entry->message->ttl);
    ttl = ttl*2 + randomi(4*TTL_DECREMENT);
    if (ttl > 4 * TTL_DECREMENT * (int) log(connectedPeers))
      ttl = 4 * TTL_DECREMENT * ((int)log(connectedPeers)) + 
	randomi(2*TTL_DECREMENT);
  }
  entry->lasttime = now;
  if (NO == checkAnonymityPolicy(AFS_CS_PROTO_QUERY,
				 ntohs(entry->message->header.size)+sizeof(HostIdentity))) {
    LOG(LOG_DEBUG,
	"DEBUG: not sending query %s due to anonymity policy!\n",
	&hex);
    return; 
  }
  IFLOG(LOG_DEBUG,
	hash2hex(&entry->message->queries[0],
		 &hex));
  entry->message->ttl = htonl(ttl);
  entry->message->priority 
    = htonl(ntohl(entry->message->priority)+
	    randomi(entry->tries + ntohl(entry->message->priority)));
  LOG(LOG_DEBUG,
      "DEBUG: sending query %s with ttl %d and priority %u (%d)\n",
      &hex,
      ttl,
      ntohl(entry->message->priority),
      this->initialTTL);

  entry->tries++;
  /* If retrying, tell client new stats */
  if(entry->tries > 1) {
    NodeContext * nc;
  
    nc=(NodeContext *)entry->data;
    nc->stats.totalRetries++;
    nc->stats.currentRetries++;

    nc->stats.requestsPending = this->requestListIndex;
    nc->stats.requestsSent = this->requestedListIndex;
    nc->stats.currentTTL = ttl;
    nc->stats.duplicationEstimate = this->duplicationEstimate;
    
    /* But only now and then */
    if((nc->stats.totalRetries % 2) == 0) {
      nc->pmodel(&nc->stats, nc->data);
    }
  }

  if (0 == (entry->tries % (MAX_TRIES * 50))) {
    IFLOG(LOG_WARNING,
	  hash2hex(&entry->message->queries[0],
		   &hex));
    LOG(LOG_WARNING,
	"WARNING: %s seems to be not available on the network\n",
	&hex);
    entry->receiverNode->print(entry->receiverNode, 0);
  }
  msg = MALLOC(ntohs(entry->message->header.size));
  memcpy(msg,
	 entry->message,
	 ntohs(entry->message->header.size));
  sock = this->sock;
  MUTEX_UNLOCK(&this->lock);
  /* writeToSocket may block if the write queue is full; if
     we at that point hold the lock, our receive thread may
     also block on reading the result from gnunetd, resulting
     in a really nasty deadlock; avoided by copying the message
     and doing the write call without holding the lock. */
  if (sock != NULL) /* destroyRM may set this->sock to NULL at ANY point! */
    if (SYSERR == writeToSocket(sock,
				msg)) 
      LOG(LOG_WARNING,
	  "WARNING: could not send request to gnunetd, "\
	  "is it running?\n");
  FREE(msg);
  MUTEX_LOCK(&this->lock);
}

/**
 * Cron job that re-issues requests. Should compute how long to sleep
 * (min ttl until next job is ready) and re-schedule itself
 * accordingly!
 **/
static void requestJob(RequestManager * this) {
  cron_t minSleep;
  cron_t now;
  cron_t delta;
  int i;  

  minSleep = -1;
  LOG(LOG_CRON,
      "CRON: requestJob %x running\n",
      (int)this);
  MUTEX_LOCK(&this->lock);
  cronTime(&now);
  for (i=0;i<this->requestedListIndex;i++) {
    delta = (this->requestedList[i]->lasttime +
	     ntohl(this->requestedList[i]->message->ttl)) - now;
    if ((this->requestedList[i]->lasttime +
	 ntohl(this->requestedList[i]->message->ttl)) <= now) {
      delta = ntohl(this->requestedList[i]->message->ttl)*2;    
      issueRequest(this, i);
    } else
      delta = (this->requestedList[i]->lasttime +
	       ntohl(this->requestedList[i]->message->ttl)) - now;
      
    if ( (delta < minSleep) || (minSleep == -1) )
      minSleep = delta;
  }
  if (minSleep < cronMILLIS * 10)
    minSleep = cronMILLIS * 10; /* maximum resolution: 10ms */
  if (this->requestedListIndex > 0) {
    if (minSleep == -1) 
      errexit("FATAL: assert failed, minsleep should be >0 "\
	      "if there are still jobs pending!\n");
    LOG(LOG_CRON,
	"CRON: scheduling next run for in %dms\n",
	minSleep);
    addCronJob((CronJob)&requestJob, 
	       minSleep, 
	       0, 
	       this);
  } else
    LOG(LOG_DEBUG,
	"DEBUG: no more jobs pending, cron not renewed!\n");
  MUTEX_UNLOCK(&this->lock);
}

/**
 * Listen on socket and receive messages. Call requestManagerReceive
 * on every reply. Never returns, if the RM dies, it will cancel us.
 **/
static void receiveThread(RequestManager * this) {
  CS_HEADER * buffer;
  GNUNET_TCP_SOCKET * sock;

  while (this->sock != NULL) {
    MUTEX_LOCK(&this->lock);
    sock = this->sock;
    MUTEX_UNLOCK(&this->lock);
    if (sock == NULL)
      break;
    buffer = NULL;
    if (SYSERR == readFromSocket(sock,
				 &buffer) ) {
      if (this->sock == NULL)
	break;
      LOG(LOG_WARNING,
	  "WARNING: requestmanager-receive thread could not "\
	  "read data from gnunetd, is the server running?\n");
      sleep(15);
      continue;
    }
    if ( (ntohs(buffer->tcpType) == AFS_CS_PROTO_RESULT_CHK) &&
	 (ntohs(buffer->size) == sizeof(AFS_CS_RESULT_CHK) ) ) {
      requestManagerReceive(this, 
			    (AFS_CS_RESULT_CHK*) buffer);
    } else {
      /* This should no longer happen, but better check than sorry... */
      LOG(LOG_WARNING,
	  "WARNING: received unexpected message (%d) from gnunetd. "
	  "(this is a bug, though we can probably recover gracefully)\n",
	  ntohs(buffer->tcpType));
      MUTEX_LOCK(&this->lock);
      releaseClientSocket(this->sock);
      this->sock = getClientSocket();
      MUTEX_UNLOCK(&this->lock);
    }
    FREE(buffer);
  }
  SEMAPHORE_UP(this->destroySignal);
}

/**
 * Create a request manager. Will create the request manager
 * datastructures and also connect to gnunetd. Creates thread that
 * listens to gnunetd replies and another thread that periodically
 * re-issues the queries. Use destroyRequestManager to abort and/or to
 * free resources after the download is complete. The callback method
 * in nc will be invoked to notify the caller of the download progress
 * such that it is possible to tell when we are done.
 *
 * @return NULL on error
 **/
RequestManager * createRequestManager() {
  RequestManager * rm;
  pthread_t receiveThread_;

  rm = MALLOC(sizeof(RequestManager));
  MUTEX_CREATE_RECURSIVE(&rm->lock);
  rm->requestListIndex 
    = 0;
  rm->requestListSize  
    = 0;
  rm->requestList      
    = NULL;
  GROW(rm->requestList,
       rm->requestListSize,
       32);
  rm->requestedListIndex
    = 0;
  memset(&rm->requestedList,
	 0,
	 sizeof(RequestEntry*) * MAX_REQUESTS);
  rm->initialTTL
    = 5 * cronSECONDS;
  rm->duplicationEstimate
    = 0;
  rm->sock
    = getClientSocket();
  rm->destroySignal
    = SEMAPHORE_NEW(0);
  if (rm->sock == NULL) {
    LOG(LOG_WARNING,
	"WARNING: could not create socket to connect to gnunetd\n");
    SEMAPHORE_UP(rm->destroySignal);
    destroyRequestManager(rm);
    return NULL;
  }
  queryPeerCount(rm->sock);
  if (0 != PTHREAD_CREATE(&receiveThread_,
			  (void*(*)(void*))&receiveThread,
			  rm, 
			  256*1024)) {
    LOG(LOG_ERROR,
	    "ERROR: could not create receiveThread (%s)\n",
	  strerror(errno));
    SEMAPHORE_UP(rm->destroySignal);
    destroyRequestManager(rm);
    return NULL;
  }
  pthread_detach(receiveThread_);
  return rm;
}

/**
 * Destroy the resources associated with a request manager.
 * Invoke this method to abort the download or to clean up
 * after the download is complete.
 *
 * @param this the request manager struct from createRequestManager
 **/
void destroyRequestManager(RequestManager * this) {
  GNUNET_TCP_SOCKET * sock;
  int i;

  suspendCron();
  MUTEX_LOCK(&this->lock);
  sock = this->sock;
  this->sock = NULL;
  MUTEX_UNLOCK(&this->lock);
  destroySocket(sock); /* destroying the socket will
			  cause the read to fail, unblocking
			  the reqeust-manager thread */
  delCronJob((CronJob)&requestJob, 
	     0, 
	     this);
  SEMAPHORE_DOWN(this->destroySignal);
  SEMAPHORE_FREE(this->destroySignal);
  MUTEX_DESTROY(&this->lock);
  for (i=0;i<this->requestListIndex;i++) 
    FREENONNULL(this->requestList[i]);
  for (i=0;i<this->requestedListIndex;i++)
    FREENONNULL(this->requestedList[i]);
  GROW(this->requestList,
       this->requestListSize,
       0);
  FREE(this);
  releaseClientSocket(sock);
  resumeCron();
}

/**
 * This method receives data corresponding to the indicated filename
 * (hashcode). Finds the Listener that scheduled this request and drop
 * it from the list of pending requests.<p>
 * 
 * @param this the request manager struct from createRequestManager
 * @param msg the message received from gnunetd
 **/ 
void requestManagerReceive(RequestManager * this,
			   AFS_CS_RESULT_CHK * msg) {
  int pos;
  int i;
  int j;
  HashCode160 query;
  RequestEntry * entry;

  /* check type of reply msg, fill in query */
  hash(&msg->result,
       sizeof(CONTENT_Block),
       &query);  
  suspendCron();
  MUTEX_LOCK(&this->lock);
  pos = -1;  
  /* find which query matches the reply, call the callback
     and recycle the slot */
  for (i=0;i<this->requestedListIndex;i++) {
    AFS_CS_QUERY * acq;

    acq = this->requestedList[i]->message;
    j = (ntohs(acq->header.size)-sizeof(AFS_CS_QUERY))/sizeof(HashCode160);    
    while ( j > 0 ) {
      j--;
      if (equalsHashCode160(&query, 
			    &acq->queries[j])) 
	pos = i;
    }
  }
  if (pos == -1) {
    this->duplicationEstimate++;
    LOG(LOG_DEBUG, 
	"DEBUG: RequestManager: received useless data (%d, %ds)!\n",
	this->duplicationEstimate,
	this->initialTTL / cronSECONDS);   
    MUTEX_UNLOCK(&this->lock);
    resumeCron();
    return;
  } 
  entry = this->requestedList[pos];
  
  if ( (entry->lasttime < cronTime(NULL)) &&
       (entry->lasttime != 0) ) /* we can get replies for things that we never
				   asked for, do not count the ttl for those!
				   Also, some ttls are adjusted into the future, they
				   should also not count! */
    this->initialTTL = ((3 * this->initialTTL) + (cronTime(NULL)-entry->lasttime)) / 4;
  /* and finally use the entry to notify the node
     that we got a reply! */
  LOG(LOG_DEBUG,
      "DEBUG: request manager receives data for %x\n",
      entry->receiverNode);
  if (SYSERR == entry->receiver(entry->receiverNode,
				&query,
				msg,
				this,
				entry->data)) {
    /* ABORT download, receiver method has
       already notified the controller via
       the pmodel callback, we need to stop
       requesting... */
    this->requestedListIndex = 0;
    this->requestListIndex = 0;
    delCronJob((CronJob)&requestJob,
	       0, 
	       this);   
  }
  MUTEX_UNLOCK(&this->lock);
  resumeCron();
}

/**
 * Queue a request for execution.
 * 
 * @param this the request manager struct from createRequestManager
 * @param node the node to call once a reply is received
 * @param callback the method to invoke
 * @param data the data argument to the Listener
 * @param message the query to send to gnunetd, freed by callee!
 **/
void requestManagerRequest(RequestManager * this,
			   Block * node,
			   Listener callback,
			   void * data,
			   AFS_CS_QUERY * message) {
  RequestEntry * entry;

  LOG(LOG_DEBUG,
      "DEBUG: requestManagerRequest for %x with callback %x\n",
      node,
      callback);
  entry = MALLOC(sizeof(RequestEntry));
  entry->message 
    = message;
  entry->lasttime 
    = 0; /* never sent */
  entry->receiver 
    = callback;
  entry->receiverNode
    = node;
  entry->data 
    = data;
  entry->tries 
    = 0; /* not tried so far */

  suspendCron();
  MUTEX_LOCK(&this->lock);   
  /* can we add to current list & issue instantly? */
  if (this->requestedListIndex < MAX_REQUESTS) {
    this->requestedList[this->requestedListIndex++] = entry;
    delCronJob((CronJob)&requestJob,
	       0, 
	       this);
    LOG(LOG_CRON,
	"CRON: scheduling next run for now!\n");
    addCronJob((CronJob)&requestJob,
	       0, 
	       0,
	       this); 
    MUTEX_UNLOCK(&this->lock);
    resumeCron();
    return;
  }
  if (this->requestListSize == this->requestListIndex) 
    GROW(this->requestList,
	 this->requestListSize,
	 this->requestListSize*2);
  /* append to the queue */
  this->requestList[this->requestListIndex++] = entry;
  MUTEX_UNLOCK(&this->lock);
  resumeCron();
}

/**
 * Assert that there are no pending requests for this node.
 **/
void requestManagerAssertDead(RequestManager * this,
			      Block * node) {
  int i;

  if (this == NULL)
    return; /* do not check */
  MUTEX_LOCK(&this->lock);
  for (i=0;i<this->requestedListIndex;i++) 
    if (this->requestedList[i]->receiverNode == node) 
      errexit("FATAL: node %x is being destroyed while request is pending\n",
	      node);
  for (i=0;i<this->requestListIndex;i++)
    if (this->requestList[i]->receiverNode == node) 
      errexit("FATAL: node %x is being destroyed while request is pending\n",
	      node);
  MUTEX_UNLOCK(&this->lock);
}

/**
 * Update a request. This method is used to selectively
 * change a query or drop it entirely.
 *
 * @param this the request manager struct from createRequestManager
 * @param node the block for which the request is updated
 * @param msg the new query message for that node, NULL for 
 *        none (then the request is dropped)
 **/
void requestManagerUpdate(RequestManager * this,
			  Block * node,
			  AFS_CS_QUERY * msg) {
  int i;

  LOG(LOG_DEBUG,
      "DEBUG: updating request for %x to %x\n",
      node, 
      msg);
  MUTEX_LOCK(&this->lock);
  for (i=0;i<this->requestedListIndex;i++) {
    if (this->requestedList[i]->receiverNode == node) {
      if (msg != NULL) { /* update */
	msg->ttl 
	  = this->requestedList[i]->message->ttl;
	msg->priority 
	  = this->requestedList[i]->message->priority;
	FREE(this->requestedList[i]->message);
	this->requestedList[i]->message = msg;	
	/* also wait a bit longer before re-issueing the
	   request, after all, we got at least one of the
	   replies! */
	this->requestedList[i]->lasttime 
	  = cronTime(NULL) + cronSECONDS; /* add 1s grace time if we got a reply
					     to a multi-query; this dramatically
					     improves the amount of "useless"
					     (duplicate) replies we get! */
      } else { /* delete */
        /* update stats */
        if(this->requestedList[i]->tries > 1)
          ((NodeContext *)(this->requestedList[i]->data))->stats.currentRetries 
	     -= (this->requestedList[i]->tries - 1);
	FREE(this->requestedList[i]->message);
	FREE(this->requestedList[i]);
	if (this->requestListIndex > 0) {
	  this->requestedList[i] 
	    = this->requestList[--this->requestListIndex];
	  this->requestList[this->requestListIndex] = NULL;
	} else {
	  this->requestedList[i] 
	    = this->requestedList[--this->requestedListIndex];
	  this->requestedList[this->requestedListIndex] 
	    = NULL;
	}
      }
      MUTEX_UNLOCK(&this->lock);
      return; /* found! */
    }
  }
  for (i=0;i<this->requestListIndex;i++)
    if (this->requestList[i]->receiverNode == node) {
      /* update stats */
      if(this->requestList[i]->tries > 1)
        ((NodeContext *)(this->requestList[i]->data))->stats.currentRetries 
		-= (this->requestList[i]->tries - 1);
      FREE(this->requestList[i]->message);
      FREE(this->requestList[i]);
      this->requestList[i] 
	= this->requestList[--this->requestListIndex];
      this->requestList[this->requestListIndex] 
	= NULL;
      MUTEX_UNLOCK(&this->lock);
      return; /* found! */
    }
  MUTEX_UNLOCK(&this->lock);  
  breakpoint();
  LOG(LOG_DEBUG,
      "DEBUG: rm failed to cancel request for %x\n",
      node);
}

/* end of requestmanager.c */
