/*
     This file is part of GNUnet

     GNUnet is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     GNUnet is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with GNUnet; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/

/**
 * @file src/applications/afs/gtkui/insert.c
 * @author Igor Wronsky
 *
 * This handles file insertions in the GUI
 **/

#include "config.h"

#include <gtk/gtk.h>
#include <string.h>

#include "util/checksum.h"
#include "util/configuration.h"
#include "util/returnvalues.h"
#include "util/xmalloc.h"
#include "applications/afs/encoding/insertutil.h"
#include "applications/afs/encoding/keywords.h"
#include "applications/afs/gtkui/helper.h"
#include "applications/afs/gtkui/insert.h"
#include "applications/afs/gtkui/main.h"
#include "util/port.h"

static void startInsert(GtkWidget * dummy, 
                        InsertModel * ilm);
static void insertModelCallback(ProgressStats * stats,
				InsertModel * ilm);

/**
 * Kill insert thread 
 *
 * Notes: The order of destruction is a bit intricate here. Destroying
 * any window kills the previous window, which kills a previous window,
 * etc, until the last one calls this function.
 *
 * (The windows will appear in order selectFileWindow, editAttributesWindow,
 *  progressBarWindow)
 *
 * @param ilm Data related to the insertion
 **/
static void destroy(GtkWidget * widget,
		    InsertModel *ilm) { 
  pthread_cancel(ilm->insertThread);
  FREE(ilm);
}

/**
 * The keyword add button was clicked. Add whatever 
 * is in the keyword box to the list of keywords.
 *
 * @param ilm Data related to the insertion
 **/
static void button_add_clicked(GtkWidget *w, 
			       InsertModel *ilm) {
  gchar *newKeyword;
  int len,i;

  newKeyword
    = STRDUP(gtk_entry_get_text(GTK_ENTRY(ilm->keywordLine)));
  /* Remove trailing & heading spaces */
  len=strlen(newKeyword);
  for(i=len-1;(newKeyword[i]==' ') && (i>=0);i--)
    newKeyword[i]=0;
  while(*newKeyword==' ')
    newKeyword++;
  if(newKeyword!=NULL && newKeyword[0]!=0) {
    gtk_clist_append(GTK_CLIST(ilm->keywordList),
		&newKeyword);
  } 
  gtk_entry_set_text(GTK_ENTRY(ilm->keywordLine),"");
}

/**
 * The keyword delete button was clicked. Delete the 
 * currently selected keyword.
 *
 * @param clist keyword list
 **/
static void button_del_clicked(GtkWidget *w, 
			       GtkWidget *clist) {
  GList *tmp;

  tmp=GTK_CLIST(clist)->selection;
  if(!tmp)
    return;
  
  gtk_clist_remove(GTK_CLIST(clist),(int)tmp->data);

}

/**
 * Show user a window to edit information related to this file.
 * After user is done, call startInsert.
 *
 * @param ilm Data related to the insertion
 **/
static void editAttributes(InsertModel * ilm) {
  GtkWidget * window;
  GtkWidget * vbox, * hbox;
  GtkWidget * clist, * scrolled_window;
  GtkWidget * label, * separator; 
  GtkWidget * button_add, * button_delete;
  GtkWidget * button_ok, * button_cancel;
  GtkWidget * keyword_line;
  GSList *group;
  GtkWidget * indexbutton1, * indexbutton2;
  gchar *titles[1] = { "Keyword(s) used" };
  int i;

  /* create new window for editing */
  window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
  gtk_widget_set_usize(GTK_WIDGET(window), 400, 400);
  gtk_window_set_title(GTK_WINDOW(window), "Edit attributes");
  vbox = gtk_vbox_new(FALSE, 0);
  gtk_container_add (GTK_CONTAINER (window), vbox);
  gtk_widget_show(vbox);

  ilm->editAttributesWindow=window;

  gtk_signal_connect (GTK_OBJECT(window),
              "delete_event",
              GTK_SIGNAL_FUNC(deleteEvent),
              ilm);
  gtk_signal_connect (GTK_OBJECT(window),
              "destroy",
              GTK_SIGNAL_FUNC(destroyWidget),
              ilm->selectFileWindow);
  gtk_container_set_border_width (GTK_CONTAINER(window), 10);

  /* Create a line to change the published filename */
  label=gtk_label_new("Published filename:");
  gtk_box_pack_start(GTK_BOX(vbox),
	     label, FALSE, FALSE, 0);
  gtk_widget_show(label);
 
  ilm->fileNameLine=gtk_entry_new();
  gtk_box_pack_start(GTK_BOX(vbox),
             ilm->fileNameLine, TRUE, TRUE, 0);
  gtk_entry_set_text(GTK_ENTRY(ilm->fileNameLine), 
		     ilm->fileNameRoot);
  gtk_widget_show(ilm->fileNameLine);
  
  /* Create a line to change the description */
  label=gtk_label_new("Description:");
  gtk_box_pack_start(GTK_BOX(vbox),
	     label, FALSE, FALSE, 0);
  gtk_widget_show(label);
  
  ilm->descriptionLine=gtk_entry_new();
  gtk_box_pack_start(GTK_BOX(vbox),
             ilm->descriptionLine, TRUE, TRUE, 0);
  gtk_entry_set_text(GTK_ENTRY(ilm->descriptionLine), ilm->description);
  gtk_widget_show(ilm->descriptionLine);
  
  /* add buttons to select the insertion method */
  label=gtk_label_new("Insertion method:");
  gtk_box_pack_start(GTK_BOX(vbox),
	     label, FALSE, FALSE, 0);
  gtk_widget_show(label);
  
  hbox=gtk_hbox_new(FALSE,0);
  gtk_box_pack_start(GTK_BOX(vbox), 
	hbox, TRUE, TRUE, 0);
  gtk_widget_show(hbox);

  indexbutton1 = gtk_radio_button_new_with_label (NULL, "Index only");
  ilm->indexButton=indexbutton1;
  gtk_box_pack_start (GTK_BOX (hbox), indexbutton1, TRUE, TRUE, 0);
  gtk_widget_show(indexbutton1);
  
  group = gtk_radio_button_group (GTK_RADIO_BUTTON (indexbutton1));
  indexbutton2 = gtk_radio_button_new_with_label(group, "Full insertion");
  gtk_box_pack_start (GTK_BOX (hbox), indexbutton2, TRUE, TRUE, 0);
  gtk_widget_show(indexbutton2);

  if(ilm->indexContent == YES)
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(indexbutton1), TRUE);
  else
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(indexbutton2), TRUE);

  /* add a list of keywords */
  scrolled_window = gtk_scrolled_window_new (NULL, NULL);
  gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolled_window),
                                  GTK_POLICY_AUTOMATIC, GTK_POLICY_ALWAYS);

  gtk_box_pack_start(GTK_BOX(vbox), scrolled_window, TRUE, TRUE, 0);
  gtk_widget_show (scrolled_window);

  clist = gtk_clist_new_with_titles( 1, titles); 
  ilm->keywordList=clist;

  gtk_container_add(GTK_CONTAINER(scrolled_window), clist);

  /* add the pre-extracted keywords to the list */
  gtk_clist_freeze(GTK_CLIST(clist));

  for(i=0;i<ilm->num_keywords;i++) {
    gtk_clist_append(GTK_CLIST(clist), &ilm->keywords[i]);
  }

  gtk_clist_thaw(GTK_CLIST(clist));
  gtk_widget_show(clist);

  /* add a line to input new keywords */
  keyword_line=gtk_entry_new();
  ilm->keywordLine=keyword_line;
  gtk_box_pack_start(GTK_BOX(vbox),
             keyword_line, FALSE,FALSE, 0);
  gtk_entry_set_text(GTK_ENTRY(keyword_line), "");
  gtk_signal_connect(GTK_OBJECT(keyword_line),"activate",
                     GTK_SIGNAL_FUNC(button_add_clicked),
                     ilm);
  gtk_widget_show(keyword_line);

  /* add the buttons to add and delete keywords */
  hbox = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, TRUE, 0);
  gtk_widget_show(hbox);

  button_add = gtk_button_new_with_label("Add keyword");
  button_delete = gtk_button_new_with_label("Delete keyword");

  gtk_box_pack_start(GTK_BOX(hbox), button_add, TRUE, TRUE, 0);
  gtk_box_pack_start(GTK_BOX(hbox), button_delete, TRUE, TRUE, 0);

  gtk_signal_connect(GTK_OBJECT(button_add), "clicked",
                GTK_SIGNAL_FUNC(button_add_clicked),
                ilm);
  gtk_signal_connect(GTK_OBJECT(button_delete), "clicked",
                GTK_SIGNAL_FUNC(button_del_clicked),
                (gpointer) clist);

  gtk_widget_show(button_add);
  gtk_widget_show(button_delete);

  /* add the insertion ok/cancel buttons */
  separator = gtk_hseparator_new();
  gtk_box_pack_start(GTK_BOX(vbox), separator, TRUE, TRUE, 0);
  gtk_widget_show(separator);

  hbox = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, TRUE, 0);
  gtk_widget_show(hbox);

  button_ok = gtk_button_new_with_label("Ok");
  button_cancel = gtk_button_new_with_label("Cancel");

  gtk_box_pack_start(GTK_BOX(hbox), button_ok, TRUE, TRUE, 0);
  gtk_box_pack_start(GTK_BOX(hbox), button_cancel, TRUE, TRUE, 0);

  gtk_signal_connect(GTK_OBJECT(button_ok), "clicked",
                GTK_SIGNAL_FUNC(startInsert),
                ilm);
  gtk_signal_connect(GTK_OBJECT(button_cancel), "clicked",
                GTK_SIGNAL_FUNC(destroyWidget),
                window);

  gtk_widget_show(button_ok);
  gtk_widget_show(button_cancel);

  /* all clear, show the window */
  gtk_widget_show(window);
}

/**
 * Callback for the file selection window. Launches the attribute
 * edit routine for the selected file, after keywords have been
 * extracted.
 *
 * @param ilm Data related to the insertion
 */
static void file_selected(GtkWidget * widget, 
			  InsertModel *ilm) 
{
  gchar *filename;
  FILE *fp;
  int i;

  filename 
    = gtk_file_selection_get_filename(GTK_FILE_SELECTION (ilm->selectFileWindow));
  if (filename == NULL) {
    guiMessage("Please select a file\n");
    gtk_widget_destroy(ilm->selectFileWindow);
    FREE(ilm);
    return;
  }
  /* get file length (for progressbar) */
  fp = fopen(filename, "r");
  if(!fp) {
    guiMessage("Unable to open %s\n", filename);
    gtk_widget_destroy(ilm->selectFileWindow);
    FREE(ilm);
    return;
  }
  fseek(fp, 0, SEEK_END);
  ilm->fileLength=ftell(fp);
  fclose(fp);

  /* fill up the filename related parts of ilm */
  ilm->fileName = STRDUP(filename);  
  /* if filename is '/home/user/foo', use 'foo' as the filenameRoot */
  ilm->fileNameRoot = NULL;
  for (i = strlen(ilm->fileName)-1;i>=0;i--) {
    if (ilm->fileName[i] == '/') {
      ilm->fileNameRoot = STRDUP(&ilm->fileName[i+1]);
      break;
    }
  }  

  /* Try to extract keywords */ 
  ilm->description = NULL;
  ilm->mimetype = NULL; 
  ilm->num_keywords = 0;
  ilm->keywords = NULL;
  extractKeywords(ilm->fileName,
		  &ilm->description,
		  &ilm->mimetype,
		  &ilm->keywords,
		  &ilm->num_keywords);

  if (ilm->description == NULL)
    ilm->description = STRDUP("No description supplied");
  if (ilm->mimetype == NULL)
    ilm->mimetype = STRDUP("unknown");
  
  if (testConfigurationString("GNUNET-INSERT",
  			      "INDEX-CONTENT",
			      "YES") == YES)
    ilm->indexContent=YES;
  else
    ilm->indexContent=NO;
   
  ilm->contentPriority=getConfigurationInt("GNUNET-INSERT",
  					   "CONTENT-PRIORITY");
					   
  /* allow the user to edit the insertion related info */
  editAttributes(ilm);

  gtk_widget_hide(ilm->selectFileWindow);
}

/**
 * A function to be run by the insert thread. Does the
 * actual insertion.
 *
 * @param ilm Collected data related to the insertion
 */
static void insertFile_(InsertModel * ilm) {
  int res;
  GNUNET_TCP_SOCKET * sock;
  Block * top;
  
  pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
  pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);

  sock = getClientSocket();
  top = insertFile(sock,
		   ilm->fileName,
		   (void (*)(ProgressStats *, void *))&insertModelCallback,
		   ilm);
  if (top != NULL) {
    res = insertRoot(sock,
		     top, 
		     ilm->description,
		     ilm->fileNameRoot,
		     ilm->mimetype,
		     ilm->num_keywords,
		     ilm->keywords);
  } else {
    res = SYSERR;
  }
  pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, NULL);
  gdk_threads_enter();
  if(res == OK) {
    HexName hex;
    HexName hex2;
    
    hash2hex(&top->chk.key,&hex);
    hash2hex(&top->chk.query, &hex2);

    guiMessage("Successfully processed %s\n%s\n%s\n%u %u\n", 
	       ilm->fileName,
	       (char*)&hex,
	       (char*)&hex2,
	       crc32N(top->data, top->len),
	       top->filesize);
    printf("Successfully processed %s -- %s %s %u %u\n", 
	   ilm->fileName,
	   (char*)&hex,
	   (char*)&hex2,
	   crc32N(top->data, top->len),
	   top->filesize);
  } else {
    guiMessage("Insertion of %s FAILED!\n", 
	       ilm->fileName);
  }
  gdk_threads_leave();
  pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
  if(top != NULL)
    top->done(top,NULL);
  releaseClientSocket(sock);
  /* insert complete */
  gtk_widget_destroy(ilm->progressBarWindow);
  FREE(ilm->keywords);
  FREE(ilm);
}

/**
 * Callback function to update the insert progressbar.
 *
 * @param stats Statistics related to insert progression
 * @param ilm Data related to the insertion
 **/
static void insertModelCallback(ProgressStats * stats,
				InsertModel * ilm) {
  if (_I_AM_DEAD == YES)
    return;
 
  pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, NULL);
  gdk_threads_enter();
  gtk_progress_set_value(GTK_PROGRESS(ilm->progressBar),
			 stats->progress);
  gdk_threads_leave();
  pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
}

/**
 * Collects the results of editAttributes, creates an insertion 
 * progressbar and launches the insertion thread.
 *
 * @param ilm Data related to the insertion
 **/
static void startInsert(GtkWidget *dummy, 
                       InsertModel *ilm) {
  GtkWidget * window;
  GtkWidget * box;
  GtkObject * adjustment;
  char ** newKeywords;
  int i;

  gtk_widget_hide(ilm->editAttributesWindow);

  /* get content indexing style */
  if(gtk_toggle_button_get_active((GtkToggleButton *)ilm->indexButton) == TRUE)
    ilm->indexContent=YES;
  else
    ilm->indexContent=NO;

  /* get the published filename */
  ilm->fileNameRoot=gtk_entry_get_text(GTK_ENTRY(ilm->fileNameLine));

  /* get the new description, if any */
  ilm->description=gtk_entry_get_text(GTK_ENTRY(ilm->descriptionLine));

  /* get new list of keywords */
  ilm->num_keywords=GTK_CLIST(ilm->keywordList)->rows;
  if(ilm->num_keywords==0) {
    guiMessage("No keywords\n");
    gtk_widget_destroy(ilm->editAttributesWindow);
    return;
  }

  newKeywords = (char**) MALLOC(ilm->num_keywords * sizeof(char*));
  for(i=0;i<ilm->num_keywords;i++) {
    gtk_clist_get_text(GTK_CLIST(ilm->keywordList),i,0,&newKeywords[i]);
  }
  ilm->keywords=newKeywords;

  /* create a new window for a progressbar */
  window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
  gtk_window_set_title(GTK_WINDOW(window), gtk_entry_get_text(GTK_ENTRY(ilm->fileNameLine)));
  box = gtk_hbox_new(FALSE, 0);
  gtk_container_add (GTK_CONTAINER (window), box);
  gtk_signal_connect (GTK_OBJECT (window),
              "delete_event",
              GTK_SIGNAL_FUNC(deleteEvent),
              ilm);
  gtk_signal_connect (GTK_OBJECT (window),
              "destroy",
              GTK_SIGNAL_FUNC(destroyWidget),
              ilm->editAttributesWindow);
  gtk_container_set_border_width (GTK_CONTAINER (window), 10);
 
  /* create the actual progressbar */
  ilm->progressBar = gtk_progress_bar_new();
  ilm->progressBarWindow = window;
  gtk_progress_set_show_text(GTK_PROGRESS(ilm->progressBar),
                 1);
  gtk_progress_set_format_string(GTK_PROGRESS(ilm->progressBar),
                 (gchar*)"%v bytes inserted");
  adjustment=gtk_adjustment_new(0,0,ilm->fileLength,1,0,0);
  gtk_progress_set_adjustment(GTK_PROGRESS(ilm->progressBar),GTK_ADJUSTMENT(adjustment));
  gtk_box_pack_start(GTK_BOX(box),
             ilm->progressBar, TRUE, TRUE, 0);
  gtk_progress_bar_set_orientation(GTK_PROGRESS_BAR(ilm->progressBar),
                   GTK_PROGRESS_LEFT_TO_RIGHT);
  gtk_widget_show(ilm->progressBar);
  gtk_widget_show(box);
 
  /* start the insert */
  PTHREAD_CREATE(&ilm->insertThread,
		 (void * (*)(void *)) insertFile_,
		 ilm,
		 16 * 1024);
  pthread_detach(ilm->insertThread);
  gtk_widget_show(window);
}

/**
 * Pops up a file selector for the user.
 *
 * Explanation: In insertion, functions will be called in the
 * following order,
 *
 * openSelectFile[OK click->]file_selected->editAttributes[OK click]
 * ->startInsert->newthread.insertFile_
 *
 * Notes: InsertModel "ilm" is used as a general storage for many
 * different things needed during the insertion. Its filled "here and 
 * there", the only sense being that anything used from it should be
 * initialized beforehand. 
 **/
void openSelectFile(void) {
  GtkWidget *window;
  InsertModel *ilm;

  ilm=(InsertModel *)MALLOC(sizeof(InsertModel));

  window = gtk_file_selection_new("Choose file to be inserted");
  
  gtk_signal_connect (GTK_OBJECT (window), "destroy",
                      GTK_SIGNAL_FUNC(destroy),
	              ilm);
  gtk_signal_connect (GTK_OBJECT (GTK_FILE_SELECTION (window)->ok_button),
                     "clicked", 
		      GTK_SIGNAL_FUNC(file_selected),
		      ilm);
  gtk_signal_connect (GTK_OBJECT (GTK_FILE_SELECTION (window)->cancel_button),
                     "clicked", 
		      GTK_SIGNAL_FUNC(destroyWidget),
		      window);
  
  ilm->selectFileWindow=window;

  gtk_widget_show(window);

}

/* end of insert.c */
